
import os
import pickle
import unittest
from io import BytesIO, StringIO
from os import DirEntry
from unittest.mock import MagicMock, Mock, patch

from dscan.models.scanner import (STATUS, Context, DiscoveryStage, File,
                                  ServerConfig, Stage, Task)


class FileSystemMockTestCase(unittest.TestCase):

    @staticmethod
    def build_mock(data):
        def read(size=-1):
            if size:
                if isinstance(data, BytesIO):
                    return data.getvalue()
                else:
                    return data.read(size)
            else:
                return data.read()

        handle = MagicMock(spect=open)
        handle.__enter__.return_value = handle
        handle.__exit__.return_value = False
        handle.__iter__.side_effect = data.__iter__
        handle.__next__.side_effect = data.__next__
        handle.readline = data.readline
        handle.read = read
        handle.seek = data.seek
        handle.tell = data.tell
        handle.write = data.write
        handle.writelines = data.writelines
        return handle

    def setUp(self) -> None:
        mos_isfile = patch('os.path.isfile')
        mos_stat = patch("os.stat")
        mos_access = patch('os.access')
        mos_access.return_value = True
        mos_isfile.return_value = True
        mos_isfile.start()
        mos_access.start()
        mos_stat = mos_stat.start()
        mos_stat.return_value = Mock(st_size=35)
        self.addCleanup(mos_isfile.stop)
        self.addCleanup(mos_access.stop)
        self.addCleanup(mos_stat.stop)


class TestStateFullFile(FileSystemMockTestCase):

    def setUp(self) -> None:
        super(TestStateFullFile, self).setUp()
        mopen = patch('builtins.open', spec=open)
        self.file_mock = mopen.start()
        self.file_mock.return_value = self.build_mock(
            StringIO("192.168.10.0/24\n192.168.12.0/24\n10.100.1.0/24\n"))
        self.targets_file = File('fake/run/targets.work')
        self.targets_file.open()
        self.addCleanup(mopen.stop)

    def test_open(self):
        self.assertTrue(True, self.targets_file.exists())
        self.assertTrue(True, self.targets_file.readable())
        self.assertEqual(3, self.targets_file.nlines)
        self.assertEqual(3, len(self.targets_file))
        line = self.targets_file.readline()
        self.assertEqual('192.168.10.0/24', line)
        self.assertEqual(1, self.targets_file.lineno)
        self.assertEqual(len('192.168.10.0/24\n'), self.targets_file.loc)
        self.targets_file.close()

    def test_serialize(self):
        line = self.targets_file.readline()
        self.assertEqual('192.168.10.0/24', line)
        self.assertEqual(1, self.targets_file.lineno)
        nobj_file = pickle.loads(pickle.dumps(self.targets_file))
        line = nobj_file.readline()
        self.assertEqual("192.168.12.0/24", line)
        self.assertEqual(2, nobj_file.lineno)


class TestTasks(unittest.TestCase):

    def test_curd(self):
        stage_name = "discovery"
        options = "-sS -n -p22"
        target = "192.168.0.0/24"
        task = Task(stage_name, options, target)

        self.assertEqual(stage_name, task.stage_name)
        self.assertEqual(options, task.options)
        self.assertEqual(target, task.target)
        self.assertEqual(STATUS.SCHEDULED, task.status)
        task.update(STATUS.RUNNING)
        self.assertEqual(STATUS.RUNNING, task.status)

    def test_invalid_status(self):
        stage_name = "discovery"
        options = "-sS -n -p22"
        target = "192.168.0.0/24"
        task = Task(stage_name, options, target)
        with self.assertRaises(AssertionError):
            task.update("FU")


class TestRuntimeContext(FileSystemMockTestCase):

    def setUp(self) -> None:
        outdir = "fake/reports"
        self.data_path = os.path.join(os.path.dirname(__file__), 'data')

        with open(os.path.join(self.data_path,
                               "discovery-nonstandar.xml")) as f:
            self.report1 = StringIO(f.read())

        with open(os.path.join(self.data_path, "discovery-nonstandard.xml"))\
                as f:
            self.report2 = StringIO(f.read())

        patch_scandir = patch('os.scandir')
        scandir_mock = patch_scandir.start()
        mock_direntry1 = Mock(spec=DirEntry)
        mock_direntry1.name = "discovery-nonstandar.xml"
        mock_direntry1.path = os.path.join(outdir, "discovery-nonstandar.xml")
        mock_direntry2 = Mock(spec=DirEntry)
        mock_direntry2.name = "discovery-nonstandard.xml"
        mock_direntry2.path = os.path.join(outdir, "discovery-nonstandard.xml")
        scandir_mock.return_value = iter([mock_direntry1, mock_direntry2])

        super(TestRuntimeContext, self).setUp()
        targets_path = "fake/run/targets.work"
        options = "-sS -n -p22"

        ltargets_path = "fake/run/live_hosts.work"
        resume_path = "fake/run/current.trace"

        self.mock_server_config = MagicMock(spect=ServerConfig)
        self.mock_server_config.rundir = "fake/run"
        self.mock_server_config.queue_path = targets_path
        self.mock_server_config.ltargets_path = ltargets_path
        self.mock_server_config.resume_path = resume_path
        self.mock_server_config.stage_list = [
            DiscoveryStage(targets_path, options, outdir, ltargets_path),
            Stage("stage1", ltargets_path, options, outdir),
            Stage("stage2", ltargets_path, options, outdir)
        ]
        self.mock_server_config.save_context = ServerConfig.save_context
        self.mock_server_config.outdir = outdir
        self.addCleanup(patch_scandir.stop)
        self.live_targets = StringIO()
        self.resume = BytesIO()
        self.targets = StringIO('172.16.71.132\n172.16.71.133\n')
        self.mocks = {
            "fake/run/targets.work": self.targets,
            "fake/reports/discovery-nonstandar.xml": self.report1,
            "fake/reports/discovery-nonstandard.xml": self.report2,
            "fake/run/live_hosts.work": self.live_targets,
            "fake/run/current.trace": self.resume
        }

    def side_effect(self, *args):
        name = args[0]
        if name == "fake/run/live_hosts.work" or name \
                == 'fake/run/current.trace':
            self.live_targets.seek(0)
        return self.build_mock(self.mocks[name])

    def check_tasks(self, task_data, task, stage_name, nactive,
                    expected_target, expected_nactive=1):
        target, options = task_data
        self.assertIsNotNone(task_data)
        self.assertEqual(stage_name, task.stage_name)
        self.assertEqual(expected_target, target)
        self.assertEqual("-sS -n -p22", options)
        self.assertEqual(STATUS.SCHEDULED, task.status)
        self.assertEqual(expected_nactive, nactive)

    @patch('builtins.open', spec=open)
    def test_context_full_flow(self, mock_file):
        mock_file.side_effect = self.side_effect
        context = Context(self.mock_server_config)

        self.assertEqual(3, len(context.stage_list))
        expected_active_status = [('discovery', 2, 2, "100.00%"),
                                  ('stage1', 1, 1, "100.00%"),
                                  ('stage2', 1, 0, "0.00%")]
        expected_task_status = [
            ('127.0.0.1:1010', 'stage1', 'SCHEDULED', '172.16.71.132/31'),
            ('127.0.0.2:1010', 'stage2', 'SCHEDULED', '172.16.71.132/31')
        ]

        agent = "127.0.0.1:1010"
        # starts with stage 1 discovery
        task_data = context.pop(agent)
        task = context.active.get(agent)
        nactive = len(context.active)
        self.check_tasks(task_data, task, "discovery", nactive,
                         "172.16.71.132")

        context.completed(agent)

        # stage 1 discovery target 2
        task_data = context.pop(agent)
        task = context.active.get(agent)
        nactive = len(context.active)
        self.check_tasks(task_data, task, "discovery", nactive,
                         "172.16.71.133")

        context.completed(agent)

        # stage 1
        task_data = context.pop(agent)
        task = context.active.get(agent)
        nactive = len(context.active)
        self.check_tasks(task_data, task, "stage1", nactive,
                         "172.16.71.132/31")
        # if another agent pulls now should be 2 active and stage 2 should
        # start
        task_data = context.pop("127.0.0.2:1010")
        task = context.active.get("127.0.0.2:1010")
        nactive = len(context.active)
        self.check_tasks(task_data, task, "stage2", nactive,
                         "172.16.71.132/31", 2)
        task_status = context.tasks_status()

        self.assertEqual(expected_task_status, task_status)
        context.completed(agent)

        nstages, pending, completion = context.ctx_status()[0]
        self.assertEqual(3, nstages)
        self.assertEqual(0, pending)
        self.assertEqual("66.67%", completion)
        active_status = context.active_stages_status()
        self.assertEqual(expected_active_status, active_status)
        context.completed("127.0.0.2:1010")
        nstages, pending, completion = context.ctx_status()[0]
        self.assertEqual(3, nstages)
        self.assertEqual(0, pending)
        self.assertEqual("100.00%", completion)

        self.assertIsNone(context.pop(agent))

    @patch('builtins.open', spec=open)
    def test_discovery_inc_stage(self, mock_file):
        """
        The discovery stage needs to finish to produce the list of live
        targets for the other stages.
        """
        mock_file.side_effect = self.side_effect
        context = Context(self.mock_server_config)
        self.assertEqual(3, len(context.stage_list))
        context.pop("127.0.0.1:1010")
        context.pop("127.0.0.2:1010")
        self.assertIsNone(context.pop("127.0.0.3:1010"))

    @patch('builtins.open', spec=open)
    def test_interruption(self, mock_file):
        mock_file.side_effect = self.side_effect
        agent = "127.0.0.1:1010"
        context = Context(self.mock_server_config)
        _ = context.pop(agent)
        context.interrupted(agent)

        self.assertEqual(1, len(context.pending))
        _ = context.pop(agent)
        self.assertEqual(0, len(context.pending))

    @patch('builtins.open', spec=open)
    def test_status_update(self, mock_file):
        mock_file.side_effect = self.side_effect
        agent = "127.0.0.1:1010"
        context = Context(self.mock_server_config)
        _ = context.pop(agent)
        # the completed status is short lived as it's deleted after the
        # task is updated.
        test_cases = [
            ("running", STATUS.RUNNING),
            ("downloading", STATUS.DOWNLOADING),
            ("interrupted", STATUS.INTERRUPTED),
        ]

        for name, expected in test_cases:
            method = getattr(context, name)
            method(agent)
            task = context.active.get(agent)
            if name == "interrupted":
                self.assertEqual(None, task)
                self.assertEqual(1, len(context.pending))
                task = context.pending.pop(0)
            self.assertEqual(expected, task.status)

    @patch('builtins.open', spec=open)
    def test_context_resume(self, mock_file):
        mock_file.side_effect = self.side_effect
        agent1 = "127.0.0.1:1010"
        agent2 = "127.0.0.2:1010"
        context = Context(self.mock_server_config)
        task1 = context.pop(agent1)
        task2 = context.pop(agent2)
        context.running(agent1)
        self.mock_server_config.save_context(self.mock_server_config,
                                             context)
        mock_file.assert_any_call('fake/run/current.trace', "wb")

        restored_ctx = Context.create(self.mock_server_config)
        self.assertEqual(context.cstage_name, restored_ctx.cstage_name)
        self.assertEqual(len(context.stage_list), len(restored_ctx.stage_list))
        self.assertEqual(0, len(restored_ctx.active))
        self.assertEqual(context.reports_path, restored_ctx.reports_path)
        self.assertEqual(2, len(restored_ctx.pending))
        rtask1 = restored_ctx.pop(agent1)
        self.assertEqual(task1, rtask1)
        self.assertEqual(1, len(restored_ctx.active))
        rtask2 = restored_ctx.pop(agent2)
        self.assertEqual(task2, rtask2)
        self.assertEqual(2, len(restored_ctx.active))
        self.assertIsNotNone(restored_ctx._lock)
        restored_ctx.completed(agent1)
        restored_ctx.completed(agent2)
        _ = restored_ctx.pop(agent1)
        restored_ctx.completed(agent1)
        restored_ctx.completed(agent2)
        nstages, pending, completion = restored_ctx.ctx_status()[0]
        self.assertEqual(3, nstages)
        self.assertEqual(0, pending)
        self.assertEqual("66.67%", completion)
        self.assertEqual("stage1", restored_ctx.cstage_name)

    @patch("os.stat")
    def test_create(self, os_stat):
        os_stat.return_value = Mock(st_size=0)
        ctx = Context.create(self.mock_server_config)
        self.assertIsNotNone(ctx)


if __name__ == '__main__':
    unittest.main()



import io
import os
import shutil
import threading
import unittest
from argparse import Namespace

from dscan.client import Agent
from dscan.models.scanner import Config
from dscan.out import ContextDisplay
from dscan.server import AgentHandler, DScanServer
from tests import create_config, data_path, log


class TestCase(unittest.TestCase):
    def setUp(self):
        targets = io.StringIO("127.0.0.1\n")
        options_agent = Namespace(name='data/data_agent', s='127.0.0.1',
                                  p=9011,
                                  cmd='agent')
        options_srv = Namespace(name='data', b='127.0.0.1', p=9011,
                                cmd='srv', targets=targets)

        self.cfg = create_config()
        self.settings_srv = Config(self.cfg, options_srv)
        self.settings_srv.target_optimization(targets)
        self.settings_agent = Config(self.cfg, options_agent)

    def tearDown(self):
        shutil.rmtree(os.path.join(data_path, "reports"))
        shutil.rmtree(os.path.join(data_path, "run"))
        shutil.rmtree(os.path.join(data_path, "data_agent"))

    @unittest.SkipTest
    def test_client_server_integration(self):

        server = DScanServer((self.settings_srv.host, self.settings_srv.port),
                             AgentHandler, options=self.settings_srv)

        server_thread = threading.Thread(target=server.serve_forever)
        # Exit the server thread when the main thread terminates
        server_thread.daemon = True
        agent = Agent(self.settings_agent)
        try:
            server_thread.start()
            log.info(f"Server loop running in thread:{server_thread.name}")
            out = ContextDisplay(server.ctx)
            out.show()

            agent.start()
        except KeyboardInterrupt:
            print("asking for shutdown !")
            server.shutdown()
            agent.shutdown()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# encoding: utf-8

"""
scanner.py
scanner runtime models
"""

import hashlib
import os
import pickle
import threading
import itertools
from enum import Enum
from dscan import log
from dscan.models.parsers import ReportsParser, TargetOptimization
from dscan.models.structures import Status, Report
from dscan.out import Display
from libnmap.process import NmapProcess


class Config:
    """
        Runtime configurations
    """
    SSL_CERTS = (
        'certs', 'sslcert', 'sslkey', 'ciphers', 'cert-hostname'
    )

    BASE = (
        'base', 'reports',
    )

    def __init__(self, config, options):
        """
        Holds the configuration parameters
        used at runtime for both server and agent!

        :param config: configparser with the configuration
        :type config: `configparser.ConfigParser`
        :param options: argument parser `argparse.ArgumentParser`
            with the user options
        """
        self.wspace = options.name
        self.port = options.p
        self.outdir = os.path.join(options.name, config.get(
            *self.BASE))
        os.makedirs(self.outdir, exist_ok=True)
        self.config = None
        if options.cmd == 'srv':
            self.config = ServerConfig(config, options, self.outdir)
            self.sslkey = self.get_work_path(config.get(*self.SSL_CERTS[0:3:2]))
            assert os.path.isfile(
                self.sslkey), "Certificate Private key not found"
            self.ciphers = config.get(*self.SSL_CERTS[0:4:3])
        else:
            self.host = options.s
        # set cert properties

        self.sslcert = self.get_work_path(config.get(*self.SSL_CERTS[0:2:1]))
        self.srv_hostname = config.get(*self.SSL_CERTS[0:5:4])

        assert os.path.isfile(self.sslcert), "Certificate file not found"

        digest: hashlib.sha512 = hashlib.sha512()
        try:
            with open(self.sslcert, 'rt') as cert:
                digest.update(cert.read().strip().encode("utf-8"))
                self.secret_key = digest.hexdigest().encode("utf-8")
        except OSError as ex:
            log.error(f"failed to open cert file {ex}")
            raise ex

    def get_work_path(self, path):
        return os.path.join(self.wspace, path)

    def __getattr__(self, name):
        if hasattr(self.config, name):
            return getattr(self.config, name)
        else:
            raise AttributeError(f"invalid key {name}")


class ServerConfig:
    """
    Server configuration parser.
    """
    SERVER = (
        'server', 'stats', 'targets', 'live-targets', 'trace',
    )

    SCAN_CONF = 'nmap-scan'

    def __init__(self, config, options, outdir):
        """
        :param config: configparser with the configuration
        :type config: `configparser.ConfigParser`
        :param options: argument parser `argparse.ArgumentParser`
            with the user options
        """
        self.outdir = outdir
        self.rundir = os.path.join(
            options.name, config.get(*self.SERVER[0:2:1]))
        self.queue_path = os.path.join(
            options.name, config.get(*self.SERVER[0:3:2]))
        self.ltargets_path = os.path.join(
            options.name, config.get(*self.SERVER[0:4:3]))
        self.resume_path = os.path.join(
            options.name, config.get(*self.SERVER[0:5:4]))
        self.host = options.b
        os.makedirs(self.rundir, exist_ok=True)
        # init scan stages !
        self.__create_stages(dict(config.items('nmap-scan')))

    def __create_stages(self, scan_options):
        self.stage_list = []
        for name, options in scan_options.items():
            options = scan_options.get(name)
            if name == "discovery":
                self.stage_list.append(DiscoveryStage(self.queue_path,
                                                      options, self.outdir,
                                                      self.ltargets_path))
            else:
                self.stage_list.append(Stage(name, self.ltargets_path,
                                             options, self.outdir))

    def target_optimization(self, targets):
        """
        Takes a list of ip Addresses and groups all sequential ips in
        cidr notation.

        :param targets: `list` of `str`
        :type: `list` of `str`
        """
        assert targets, "Empty target list"
        if not os.path.isfile(self.resume_path):
            queue_optimization = TargetOptimization(self.queue_path)
            queue_optimization.save(targets)

    def save_context(self, ctx):
        """
        Serializes the context to resume later.

        :param ctx: instance of `Context`
        :type ctx: `Context`
        """
        log.info(f"Saving the current context {self.resume_path}")
        with open(self.resume_path, 'wb') as rfile:
            pickle.dump(ctx, rfile)


class File:
    """
    Creates a stateful file object
    allows file to restore its previous state.
    """

    def __init__(self, path):
        """
        :param path: string of path to the file to open.
        :type path: `str` File path
        """
        self._path = path
        self._fd = None
        self.nlines = 0
        self.lineno = 0
        self.loc = 0
        self.mode = 'r'

    def open(self, mode='r'):
        assert self.exists(), f"{self._path} is not a valid file"
        assert self.readable(), f"Unable to read: {self._path}"
        if not self._fd:
            self.mode = mode
            self._fd = open(self._path, mode)
            self._line_count()

    def readline(self):
        self.open()  # open the file if its not open already
        self.lineno += 1
        line = self._fd.readline()
        self.loc = self._fd.tell()
        if not line:
            return None
        return line.strip()

    def isempty(self):
        """
        Check if the file is emtpy.

        :return: `True` if the file is empty
        """
        assert self.exists(), f"{self._path} is not a valid file"
        assert self.readable(), f"Unable to read: {self._path}"
        if os.stat(self._path).st_size > 0:
            return False
        else:
            return True

    def exists(self):
        """
        Returns a Boolean if the path is a valid file.
            proxy call to `os.path.isfile`.

        :return: `True` if path is a valid file.
        """
        return os.path.isfile(self._path)

    def readable(self):
        """
        :return: `True` if file is readable.
        """
        return os.access(self._path, os.R_OK)

    def _line_count(self):
        """
        Counts the number of lines in the file.
        """
        if self.isempty():
            return 0

        lines = 0
        buf_size = 1024 * 1024
        read_f = self.read  # loop optimization
        buf = read_f(buf_size)

        while buf:
            lines += buf.count('\n')
            if not buf.endswith('\n'):
                lines += 1
            buf = read_f(buf_size)

        self.nlines = lines
        self._fd.seek(0)

    def __getattr__(self, name):
        if hasattr(self._fd, name):
            return getattr(self._fd, name)
        else:
            raise AttributeError(f"invalid key {name}")

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        log.info(f"saving file: {self._path} loc:{self.loc} state")
        state = self.__dict__.copy()
        # Remove the unpickable entries.
        del state['_fd']
        return state

    def __setstate__(self, state):
        # Restore instance attributes (i.e., _path and nlines ...).
        self.__dict__.update(state)
        fd = None
        # if the loc is 0 then we have an uninitialized stage
        # that depends on unfinished stage, will be opened when the first
        # time a target is pulled.
        if self.loc != 0 and self.exists():
            # Restore the previously opened file's state.
            log.info(f"restoring file: {self._path} loc:{self.loc} state")
            fd = open(self._path, self.mode)

            # set the file to the prev location.
            fd.seek(self.loc)
        self._fd = fd

    def __len__(self):
        self.open()
        return self.nlines

    def __str__(self):
        return f"File(path:{self._path}, nlines: {self.nlines}, lineno:" \
               f"{self.lineno}, mode:{self.mode})"


class STATUS(Enum):
    """
    Each Scan task has the following states:
    - Scheduled: the default state set when its created.
    - Running: Set after the agent has confirmed the task has started
    executing.
    - Interrupted: Set when the task is aborted by agent, or the server has
    been halted.
    - Downloading: Set when the agent notifies its ready to sent the report.
    - Completed: Set only after the report has been received successfully
    """
    SCHEDULED = 1
    RUNNING = 2
    INTERRUPTED = 3
    DOWNLOADING = 4
    COMPLETED = 5


class Task:
    """
    Representation of a scan task.
    A scan task is sent to an agent with a target and, scan options.
    Each task has the following states:
    - Scheduled: the default state set when its created.
    - Running: Set after the agent has confirmed the task has started
    executing.
    - Interrupted: Set when the task is aborted by agent, or the server has
    been halted.
    - Downloading: Set when the agent notifies its ready to sent the report.
    - Completed: Set only after the report has been received successfully.
    """

    def __init__(self, stage_name, options, target):
        """
        :param stage_name: stage name
        :param options: scan options.
        :param target: target ip address.
        :type stage_name: `str`
        :type options: `str`
        :type target: `str`
        """
        self.stage_name = stage_name
        self.options = options
        self.target = target
        self.status = STATUS.SCHEDULED

    def update(self, status):
        assert isinstance(status, STATUS)
        self.status = status

    def as_tuple(self):
        """
        returns a tuple with the target and scan options.

        :return: tuple options, target
        :rtype tuple:
        """
        return self.stage_name, self.status.name, self.target, \
            self.options

    def __str__(self):
        return f"{self.stage_name}, {self.status.name}, {self.target}" \
               f"{self.options}"


class Stage:

    def __init__(self, stage_name, targets_path, options, outdir):
        assert targets_path, "Invalid targets file Name"
        assert stage_name, "Invalid stage Name"
        self.targets_path = targets_path
        self.name = stage_name
        self.targets = File(self.targets_path)
        self.options = options
        self.reports_path = outdir
        self.ftargets = 0

    def next_task(self):
        """
        Get next target from the file.

        :return: Task.
        :rtype: `Task`
        """
        target = self.targets.readline()
        if target:
            return Task(self.name, self.options, target)
        else:
            return None

    def inc_finished(self):
        self.ftargets += 1

    def process_results(self):
        """
        Meant to be overwritten, like for example stages like ping sweep aka
        discovery.
        """
        pass

    @property
    def isfinished(self):
        """
        Returns True if the number of lines is equal to the number of
            finished targets.

        :return: bool
        :rtype: `bool`
        """
        if self.targets.nlines == self.ftargets:
            return True
        else:
            return False

    @property
    def percentage(self):
        """
        Calculates the completion of this stage.

        :return: float of of the % completion.
        :rtype: `float`
        """
        if self.ftargets > 0:
            return float(self.ftargets) / float(self.targets.nlines) * 100
        else:
            return float(0)

    def as_tuple(self):
        """
        Returns the information as tuple, nlines, finished targets and %,
        used by Display to print scanner status.

        :return: tuple of strings.
        :rtype: `tuple` of `str`.
        """
        return self.name, self.targets.nlines, self.ftargets, \
            f"{self.percentage:.2f}%"

    def close(self):
        self.targets.close()


class DiscoveryStage(Stage):

    def __init__(self, targets_path, options, outdir, ltargets_path):
        super().__init__("discovery", targets_path, options, outdir)
        self.ltargets_path = ltargets_path

    def process_results(self):
        """
        When this stage is finished the `Context` will call this method to
        create a list of live targets.
        """
        results_parser = ReportsParser(self.reports_path, 'discovery-*.xml')
        live_queue = TargetOptimization(self.ltargets_path)
        live_queue.save(results_parser.hosts_up())


class Context:
    """
    Context is a thread safe proxy like class, responsible for all the
    execution flow and inherent logic.
    Acts as a proxy between the active stages and the `stage` implementation.
    """

    def __init__(self, options):
        self.stage_list = list(options.stage_list)
        self.nstages = len(self.stage_list)
        self.cstage_name = None
        self.active_stages = {}
        self.reports_path = options.outdir
        self.active = {}
        self.pending = []
        self._lock = threading.Lock()

    def pop(self, agent):
        """
        Gets the next `Task` from the current Active Stage, if their are no
        pending `Tasks` to be executed.
        Pending tasks are tasks that are canceled or restored from a previous
        interrupted session.
        If a stage is finished (no more targets), the next stage will take
        another stage from the list until its finished.

        :param agent:
            str with ipaddress and port in ip:port format, this allows the
            server to manage multiple agents in one host.
            to run multiple clients at once.
        :return: A target to scan! `task`
        :rtype: `tuple`
        """
        with self._lock:
            task = None
            if agent in self.active:
                # This exists to make shore we don't lose targets.
                # this would be better if we knew how many tries a target
                # had faced. TODO
                log.info(f"Agent {agent} is requesting a new task with a "
                         f"task in execution sending it again!")
                task = self.active.get(agent)
                task.update(STATUS.SCHEDULED)
                return task.as_tuple()[2:]

            if len(self.pending) > 0:
                task = self.pending.pop(0)
            else:
                cstage = self.__cstage()
                if cstage:
                    task = cstage.next_task()
                    if not task:
                        # the only stage that needs to be finished
                        # to proceed is
                        # discovery as the other stages need the
                        # list of live hosts
                        if cstage.name != "discovery" or cstage.isfinished:
                            if cstage.isfinished:
                                cstage.process_results()
                                cstage.close()
                            cstage = self.__cstage(True)
                            if cstage:
                                task = cstage.next_task()

            # if we have a valid task save it in the active collection
            if task:
                self.active.update({agent: task})
                # the consumers only need scan related information...
                return task.as_tuple()[2:]

    def completed(self, agent):
        """
        Marks a agent task as complete.

        :param agent: ip:port of agent
        :type agent: `str`
        """
        self._update_task_status(agent, STATUS.COMPLETED)

    def downloading(self, agent):
        """
        Marks a agent task as Downloading, notifying that the report
        download has started.

        :param agent: ip:port of agent
        :type agent: `str`
        """
        self._update_task_status(agent, STATUS.DOWNLOADING)

    def interrupted(self, agent):
        """
        Marks a task as interrupted, when agent disconnects for example.

        :param agent: ip:port of agent
        :type agent: `str`
        """
        self._update_task_status(agent, STATUS.INTERRUPTED)

    def running(self, agent):
        """
        After the server sends a target the agent notifies the task has
        started.

        :param agent: ip:port of agent
        :type agent: `str`
        """
        self._update_task_status(agent, STATUS.RUNNING)

    def get_report(self, agent, file_name):
        """
        :param agent: str with ipaddress and port in ip:port format
        :type agent: `str`
        :param file_name: name of the file sent by the agent.
        :type file_name: `str`
        :return: file descriptor to save the scan report.
        """
        try:
            _, tstage = self.__find_task_stage(agent)
            file_name = f"{tstage.name}-{file_name}"
            report_file = open(os.path.join(self.reports_path, file_name),
                               "wb")
            return report_file
        except Exception as ex:
            log.error(f"Unable to open report for {file_name}")
            log.error(f"{ex}")
            return None

    def _update_task_status(self, agent, status):
        """
        Internal method updates  a task of a given stage status, its also
        responsible for managing the interrupted tasks.

        :param agent: str with ipaddress and port in ip:port format
        :param status: `STATUS` value to change.
        """
        with self._lock:
            task, tstage = self.__find_task_stage(agent)
            if task and tstage:
                task.update(status)
                if status == STATUS.COMPLETED:
                    tstage.inc_finished()
                    # clean the completed task
                    del self.active[agent]
                if status == status.INTERRUPTED:
                    log.info(f"Scan of {task.target} running on {agent} was "
                             f"interrupted")
                    self.pending.append(task)
                    del self.active[agent]
            else:
                log.debug(f"Agent {agent} is trying to update {status} on "
                          f"non existing task")

    def __cstage(self, force_next=False):
        """
        :param force_next: if True wil force the stage to advance one step
        defaults to False.
        :type force_next: `bool`
        :return: An instance of `stage`
        :rtype: `stage`
        """
        try:
            if not self.cstage_name or force_next:
                stage = self.stage_list.pop(0)
                self.active_stages[stage.name] = stage
                self.cstage_name = stage.name

            return self.active_stages[self.cstage_name]
        except IndexError:
            log.error("Stage list is empty !")
            return None

    def __find_task_stage(self, agent):
        """
        :param agent: str with ipaddress and port in ip:port format
        :return: tuple `Task` and `Stage` that created that task
        :rtype: `tuple`
        """
        tstage = None
        task = self.active.get(agent, None)
        if task:
            tstage = self.active_stages.get(task.stage_name, None)
        return task, tstage

    @classmethod
    def create(cls, options):
        """
        :param options: instance of `ServerConfig`
        :type options: ´ServerConfig´
        :return: instance of `Context`
        :rtype: ´Context`
        """
        rpath = options.resume_path
        if os.path.isfile(rpath) and os.stat(rpath).st_size > 0:
            log.info("Found resume file, loading...!")
            with open(options.resume_path, 'rb') as rfile:
                # i had to make this to make this testable with mocks!
                # load with file didn't work, some how!
                data = rfile.read()
                ctx = pickle.loads(data)
                return ctx
        else:
            return cls(options)

    def tasks_status(self):
        """
        :return: list of tuple of active task's status.
        :rtype: `list` of `tuple`s
        """
        data = []
        for agent, task in self.active.items():
            data.append((agent, *task.as_tuple()[:3]))
        return data

    def active_stages_status(self):
        """
        :return: list of tuples with active stages status.
        :rtype: `list` of `tuples`
        """
        data = []
        for stage in self.active_stages.values():
            data.append(stage.as_tuple())
        return data

    @property
    def is_finished(self):
        """
        :return: bool iterates the active stages and collects all the
            `finished` properties, returns `True` if all of them are true.
        """
        status = [status.isfinished for status in self.active_stages.values()]
        return all(status) and len(status) == self.nstages

    def ctx_status(self):
        stage_comp = float(0)
        for stage in self.active_stages.values():
            stage_comp += stage.percentage
        return [(self.nstages, len(self.pending), "{:.2f}%"
                 .format((stage_comp / float(self.nstages * 100) * 100)))]

    def __getstate__(self):
        with self._lock:
            log.info("saving context state")
            state = self.__dict__.copy()
            for task in state['active'].values():
                task.update(STATUS.INTERRUPTED)
                state['pending'].append(task)

            # close file descriptors on all active stages
            for active_stage in state['active_stages'].values():
                active_stage.close()

            state['active'] = {}
            # Remove the unpickable entries.
            del state['_lock']
            return state

    def __setstate__(self, state):
        # restore the previous state, needed due to the existence of non
        # serializable objects
        self.__dict__.update(state)
        log.info("Restoring context state")
        self._lock = threading.Lock()


class ScanProcess:
    """
    Used by the agent (client side), its a proxy class
    to provide an interface to interact with `libnmap` responsible
    for the actual execution.
    Its implementation allows handling, duplicate file names, with numeric
    prefix, as well as keeping status on the execution process.
    """
    TASK_HEADERS = [
        "Target", "Nª completed Scans", "Status"
    ]

    def __init__(self, output):
        """
        wrapper around `libnmap` scan execution
        :param output: str path to save the reports
        """
        self.output = output
        # current target aka ctarget is a tuple (target, options).
        self.ctarget = None
        # number of successful scans finished
        self.number_scans = 0
        self.display = Display()
        self.status = None

    def __inc(self):
        """
        increments the number_scans.
        """
        self.number_scans += 1

    def report_name(self, extension):
        """
        Checks if a report with the current target.extension exists,
        and prepends a number if it does.

        :param extension: xml, nmap.
        :return: path str path and filename, the filename will be prefixed,
            by a number if the base+extension already exists in the outdir.
        :rtype: `str`
        """
        fname = self.ctarget[0].replace('/', '-')
        path = os.path.join(self.output, f"{fname}"
                                         f".{extension}")
        exists = os.path.isfile(path)
        prefix = itertools.count()
        while exists:
            n = next(prefix)
            path = os.path.join(self.output, f"{n}-{fname}"
                                             f".{extension}")
            exists = os.path.isfile(path)
        return path

    def run(self, target, options, callback):
        """
        Executes the scan on a given target.

        :param target:
        :param options:
        :param callback: callback function to report status to the server.
        :return: report object
        :rtype: `dscan.models.structures.Report`
        """
        self.ctarget = (target, options)
        nmap_proc = None
        try:
            options = " ".join([options, f"-oN {self.report_name('nmap')}"])
            nmap_proc = NmapProcess(targets=target, options=options,
                                    safe_mode=False,
                                    event_callback=self.show_status)

            log.info("Nmap scan started Sending success status")
            callback(Status.SUCCESS)
            rc = nmap_proc.run()
            if rc == 0:
                # after finished encode and hash the contents for transfer.
                self.__inc()
                data = nmap_proc.stdout.encode("utf-8")
                report_file = self.report_name("xml")
                with open(self.report_name("xml"), "wb") as rfile:
                    rfile.write(data)
                    rfile.flush()
                digest = hashlib.sha512(data).hexdigest()
                report = Report(len(data), os.path.basename(report_file),
                                digest)
                self.print(target, 100)
                return report
            else:
                callback(Status.FAILED)
                log.error(f"Nmap Scan failed {nmap_proc.stderr}")
        except Exception as ex:
            log.error(f"something went wrong {ex}")
            callback(Status.FAILED)
        finally:
            if nmap_proc:
                nmap_proc.stop()
                # orthodox fix NmapProcess is leaving subprocess streams open.
                subproc = getattr(nmap_proc, "_NmapProcess__nmap_proc")
                if subproc:
                    subproc.stdout.close()
                    subproc.stderr.close()

    def print(self, target, progress):
        self.display.print_table(self.TASK_HEADERS,
                                 [(target, self.number_scans,
                                   progress)], clear=True)

    def show_status(self, nmapscan=None):
        """
        :param nmapscan: takes `libnmap.process.NmapProcess` instance
            to display the current status of the scan.
        :type nmapscan: `libnmap.process.NmapProcess`
        """
        if nmapscan.is_running() and nmapscan.current_task:
            ntask = nmapscan.current_task
            self.display.print_table(self.TASK_HEADERS,
                                     [(self.ctarget[0], self.number_scans,
                                       ntask.progress)], clear=True)



import hmac
import os
import ssl
import struct
import threading
import unittest
from argparse import Namespace
from socket import AF_INET, SOCK_STREAM, socket
from unittest.mock import MagicMock, mock_open, patch

from dscan.models.scanner import Config, Context
from dscan.models.structures import (Auth, ExitStatus, Ready, Report, Status,
                                     Structure)
from dscan.server import AgentHandler, DScanServer
from tests import BufMock, create_config, data_path, log


class TestAgentHandler(unittest.TestCase):

    @staticmethod
    def build_mock(data):
        handle = MagicMock(spect=open)
        handle.__enter__.return_value = handle
        handle.__exit__.return_value = False
        handle.__iter__.side_effect = data.__iter__
        handle.__next__.side_effect = data.__next__
        handle.readline = data.readline
        handle.read = data.read
        handle.seek = data.seek
        handle.tell = data.tell
        handle.write = data.write
        handle.writelines = data.writelines
        return handle

    def setUp(self):
        options = Namespace(name='data', b='127.0.0.1', p=9011,
                            cmd='srv', targets='foofile')
        self.cfg = create_config()
        self.patcher = patch('os.makedirs')
        mos_isfile = patch('os.path.isfile')
        mos_isfile.return_value = True
        mos_isfile.start()
        self.mock_makedirs = self.patcher.start()
        self.settings = Config(self.cfg, options)
        # ctx = Context(self.settings)
        self.ctx = MagicMock(spect=Context)
        self.ctx.pop.return_value = ("127.0.0.1", "-sV -Pn -p1-1000")
        self.ctx.secret_key = self.settings.secret_key
        self.ctx.is_finished = False
        self.mock_server = MagicMock(spect=DScanServer)
        self.mock_server.secret_key = self.settings.secret_key
        self.mock_terminate = MagicMock()
        self.mock_terminate.is_set.return_value = False
        self.hmac_patch = patch.object(hmac, 'compare_digest',
                                       return_value=True)
        self.challenge = b'4%x8p\x8d\xda\x04\xe5r\xfb\xc1Si8[' \
                         b'\xcb\x1a\x1c\x84\xf5\xb5%\x15[' \
                         b'\xea\x10\x96)!n\xe3\xadit\x0f\x15e\xc2\x06\xd1' \
                         b'\xd8\xb0' \
                         b'\xc5\x81\x87\xf2s\xe3\xd8\x95\xd1\x9c\xbdM\x8f' \
                         b'\x9c\xd5' \
                         b'\x14\xb3\x8e\xdd\x8eQ\xffw\x10Y8\xa5\xa5\x83\xf3' \
                         b'\xeeQ' \
                         b'\xa1\xfcOP\x9d\xd6\x80x\x80\x9eh\x11\xa7\xd7\xce' \
                         b'\xcf' \
                         b'.\xec\x01\x94S\xd4\x1d\x7f\xef\x83e\xe8\xfa\xf9' \
                         b'`\xfb' \
                         b'\xc6:SB\xeff\x15\r\xcb\xe9\xa4\xefO\x03i\xe9' \
                         b'\xefoMz\x8b'
        self.hmac_patch.start()
        self.addCleanup(self.patcher.stop)
        self.addCleanup(self.hmac_patch.stop)
        self.addCleanup(mos_isfile.stop)

    def tearDown(self):
        self.patcher.stop()
        self.hmac_patch.stop()

    @patch('socket.socket')
    @patch('os.urandom')
    def test_authentication(self, mock_urandom, mock_socket):

        mock_urandom.return_value = self.challenge
        # emulate the client!

        hmac_hash = hmac.new(self.settings.secret_key, self.challenge,
                             'sha512')
        digest = hmac_hash.hexdigest().encode("utf-8")
        buffer = BufMock(Auth(digest))
        mock_socket.recv = buffer.read
        mock = MagicMock()
        handler = AgentHandler(mock_socket, ('127.0.0.1', '1234'),
                               self.mock_server, terminate_event=self.mock_terminate,
                               context=self.ctx)
        self.assertTrue(handler.authenticated)

    @patch('socket.socket')
    def test_ready(self, mock_socket):
        buffer = BufMock(Auth(self.challenge), Ready(0, "bub"),
                         struct.pack("<B", 0))
        mock_socket.recv = buffer.read

        AgentHandler(mock_socket, ('127.0.0.1', '1234'), self.mock_server,
                     terminate_event=self.mock_terminate, context=self.ctx)
        mock_socket.sendall.assert_called_with(
            b'\x03\t\x10127.0.0.1-sV -Pn -p1-1000')
        self.ctx.running.assert_called_once()
        self.ctx.running.assert_called_with("127.0.0.1:1234")

    @patch('socket.socket')
    def test_wait(self, mock_socket):
        buffer = BufMock(Auth(self.challenge), Ready(0, "bub"),
                         Ready(0, "bub"), struct.pack("<B", 0))
        mock_socket.recv = buffer.read
        mock_data = MagicMock()
        mock_data.side_effect = [None, ("127.0.0.1", "-sV -Pn -p1-1000")]
        self.ctx.pop = mock_data

        AgentHandler(mock_socket, ('127.0.0.1', '1234'), self.mock_server,
                     terminate_event=self.mock_terminate, context=self.ctx)

        mock_socket.sendall.assert_called_with(
            b'\x03\t\x10127.0.0.1-sV -Pn -p1-1000')
        self.ctx.running.assert_called_once()
        self.ctx.running.assert_called_with("127.0.0.1:1234")

    @patch('socket.socket')
    def test_ready_failed(self, mock_socket):
        buffer = BufMock(Auth(self.challenge), Ready(0, "bub"),
                         struct.pack("<B", 1))
        mock_socket.recv = buffer.read

        AgentHandler(mock_socket, ('127.0.0.1', '1234'), self.mock_server,
                     terminate_event=self.mock_terminate, context=self.ctx)
        mock_socket.sendall.assert_called_with(
            b'\x03\t\x10127.0.0.1-sV -Pn -p1-1000')
        self.ctx.running.assert_not_called()
        self.ctx.interrupted.assert_called_once()

    @patch('socket.socket')
    def test_ready_disconnected(self, mock_socket):
        buffer = BufMock(Auth(self.challenge), Ready(0, "bub"))
        mock_socket.recv = buffer.read

        AgentHandler(mock_socket, ('127.0.0.1', '1234'), self.mock_server,
                     terminate_event=self.mock_terminate, context=self.ctx)
        mock_socket.sendall.assert_called_with(
            b'\x03\t\x10127.0.0.1-sV -Pn -p1-1000')
        self.ctx.running.assert_not_called()
        self.ctx.interrupted.assert_called_once()

    @patch('socket.socket')
    def test_report_send(self, mock_socket):
        expected_hash = "055a61499ea7c0d96332cf850f69ecb7295"

        file = open(os.path.join(data_path, 'discovery-nonstandar.xml'),
                    'rb')
        file.seek(0, os.SEEK_END)
        report_msg = Report(file.tell(), "foobar.xml", expected_hash)
        file.seek(0)
        buffer = BufMock(Auth("fu").pack(), report_msg.pack(), file)
        mock_socket.recv = buffer.read
        report_mock = mock_open()
        with patch('builtins.open', report_mock) as m:
            self.ctx.get_report = report_mock
            AgentHandler(mock_socket, ('127.0.0.1', '1234'), self.mock_server,
                         terminate_event=self.mock_terminate, context=self.ctx)
            handle = m.return_value
            offset = file.tell() - 996
            file.seek(offset)
            self.assertEqual(handle.write.call_count, 3)
            handle.write.assert_called_with(file.read())
            self.ctx.downloading.assert_called_with("127.0.0.1:1234")
            self.ctx.completed.assert_called_with("127.0.0.1:1234")
        file.close()

    @patch("builtins.open")
    def test_server(self, mock_open):
        with patch('os.path.isfile') as misfile:
            misfile.return_value = False
            server = DScanServer((self.settings.host, self.settings.port),
                                 AgentHandler, options=self.settings)

            server_thread = threading.Thread(target=server.serve_forever)
            # Exit the server thread when the main thread terminates
            server_thread.daemon = True
            server_thread.start()
            log.info(f"Server loop running in thread:{server_thread.name}")
            context = ssl.create_default_context()
            context.check_hostname = False
            context.load_verify_locations(self.settings.sslcert)
            s = context.wrap_socket(socket(AF_INET, SOCK_STREAM),
                                    server_side=False, server_hostname="dscan")
            s.connect(('127.0.0.1', 9011))

            opr = Structure.create(s)
            hmac_hash = hmac.new(self.settings.secret_key, opr.data,
                                 'sha512')
            digest = hmac_hash.hexdigest().encode("utf-8")

            s.sendall(Auth(digest).pack())
            s.close()
            server.shutdown()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# encoding: utf-8

"""
out.py
Display the information and status.
"""
import sys
import termios
import threading


class Display:
    """
    Visualisation of the current execution status.
    """
    CLEAR_SCREEN = "\033[H\033[2J\033[3J"
    CLEAR_CURRENT_LINE = "\033[1K\033[0G"
    TITLE = "Distributed Scan Status"

    def __init__(self):
        self.echo = True

    def disable_echo(self):
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        new = termios.tcgetattr(fd)
        new[3] = new[3] & ~termios.ECHO  # lflags
        # try:
        termios.tcsetattr(fd, termios.TCSADRAIN, new)
        self.echo = False
        # finally:
        # termios.tcsetattr(fd, termios.TCSADRAIN, old)

    def inline(self, message):
        """
        :param message: string message to print underneath the table printer.
        """
        if self.echo:
            self.disable_echo()
        sys.stdout.write(self.CLEAR_CURRENT_LINE)
        sys.stdout.write(f"message:\t{message}")
        sys.stdout.flush()

    def print_title(self, title):
        """
        :param title: Title of the tabular output.
        """
        if self.echo:
            self.disable_echo()
        title_length = len(title)
        sys.stdout.write("\n\n")
        sys.stdout.write("%s\n" % title.ljust(title_length))
        sys.stdout.write("%s\n\n" % ("=" * title_length))

    def print_table(self, headers, data, clear=False):
        """
        Created based on active code recepie.

        :param headers: `list` of `str` table headers.
        :param data: a `list` of `tuple` with the data to be showed.
        :param clear: `bool` clear the previous print.
        """
        if self.echo:
            self.disable_echo()
        result = []
        names = list(headers)
        result.append(names)
        result.extend(list(map(str, row)) for row in data)

        lens = [list(map(len, row)) for row in result]
        field_lens = list(map(max, zip(*lens)))
        result.insert(0, ['-' * length for length in field_lens])
        result.insert(2, ['-' * length for length in field_lens])

        format_string = "\t".join(
            '{%d:%ds}' % item for item in enumerate(field_lens))

        if clear:
            sys.stdout.write(self.CLEAR_SCREEN)
            title_length = len(self.TITLE)
            sys.stdout.write("\n\n")
            sys.stdout.write("%s\n" % self.TITLE.ljust(title_length))
            sys.stdout.write("%s\n\n" % ("=" * title_length))

        sys.stdout.write(
            '\n'.join(format_string.format(*row) for row in result))
        sys.stdout.write("\n\n")
        sys.stdout.flush()

    def show(self):
        """
        meant to be overwritten
        """
        pass


class ContextDisplay(Display):
    """
    Context is displayed as a table.
    """
    STAGES_HEADERS = [
        "Nº Stages", "Nº Pending Tasks", "Completion %"
    ]

    ACTIVE_STAGES_HEADERS = [
        'Stage', "Nº Targets", "Nº Finished", "Completion %"
    ]

    TASK_HEADERS = [
        "Agent", "Stage", "Task Status", "Target Ip"
    ]

    def __init__(self, context):
        super().__init__()
        self.ctx = context

    def show(self):
        """
        Takes the context's status, active stages status, and task status
        and prints them in a tabular format.
        """
        self.print_table(self.STAGES_HEADERS, self.ctx.ctx_status(), True)
        self.print_table(self.ACTIVE_STAGES_HEADERS,
                         self.ctx.active_stages_status())
        self.print_table(self.TASK_HEADERS, self.ctx.tasks_status())

        if not self.ctx.is_finished:
            threading.Timer(1, self.show).start()



import hashlib
import io
import struct
import unittest
from socket import socket
from unittest.mock import MagicMock, patch

from dscan.models.structures import (Auth, Command, ExitStatus, Operations,
                                     Ready, Report, Status, Structure)


class TestStructure(unittest.TestCase):

    @staticmethod
    def build_mock(data):
        data = io.BytesIO(data.pack())
        handle = MagicMock(spect=socket)
        handle.recv = data.read1
        return handle

    @staticmethod
    def getSize():
        with open(__file__, 'r') as fileobject:
            fileobject.seek(0, 2)  # move the cursor to the end of the file
            size = fileobject.tell()
        return size

    def test_auth_pack_unpack(self):
        digest = hashlib.sha512(b"pickabu").hexdigest()
        expected = Auth(digest)

        mock_sock = self.build_mock(expected)
        with patch('socket.socket', new=mock_sock) as mock_socket:
            struct_size: int = struct.calcsize("<B")
            result = Structure.create(sock=mock_socket)
            self.assertEqual(expected.data, result.data)
            self.assertEqual(digest.encode("ascii"), result.data)
            self.assertEqual(Operations.AUTH, result.op_code)

    def test_ready_pack_unpack(self):
        expected = Ready(0, 'agentx')
        mock_sock = self.build_mock(expected)
        with patch('socket.socket', new=mock_sock) as mock_socket:
            result = Structure.create(sock=mock_socket)
            self.assertEqual(expected.uid, result.uid)
            self.assertEqual(expected.alias, result.alias)

    def test_result_status_pack_unpack(self):
        expected = ExitStatus(Status.FINISHED)
        mock_sock = self.build_mock(expected)
        with patch('socket.socket', new=mock_sock) as mock_socket:
            result = Structure.create(sock=mock_socket)
            self.assertEqual(expected.op_code, result.op_code)
            self.assertEqual(expected.status, result.status)

    def test_command_pack_unpack(self):
        expected = Command("127.0.0.1", "-sV -Pn -p1-1000")
        mock_sock = self.build_mock(expected)
        with patch('socket.socket', new=mock_sock) as mock_socket:
            result = Structure.create(sock=mock_socket)
            self.assertEqual(expected.target, result.target)

    def test_report_pack_unpack(self):
        digest = hashlib.sha512(b"pickabu").hexdigest()
        expected = Report(self.getSize(), "fu.xml", digest)
        mock_sock = self.build_mock(expected)
        with patch('socket.socket', new=mock_sock) as mock_socket:
            result = Structure.create(sock=mock_socket)
            self.assertEqual(expected.filesize, result.filesize)
            self.assertEqual(expected.filename, result.filename)
            self.assertEqual(expected.filehash, result.filehash)

    def test_status(self):
        self.assertTrue((0 == Status.SUCCESS.value))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock, Mock, PropertyMock, call, patch

import tests
from dscan.models.scanner import ScanProcess, Status


class TestScanProcess(unittest.TestCase):
    def setUp(self):
        self.callback = Mock()
        self.patch_nmap_proc = patch('dscan.models.scanner.NmapProcess')
        mopen = patch('builtins.open', spec=open)
        self.file_mock = mopen.start()
        # patching sleep to speed the test execution
        running_mock = MagicMock(side_effect=[True, True, True, False])
        run_mock = MagicMock(return_value=0)
        progress_mock = PropertyMock(side_effect=[10, 50, 100])
        self.mock_nmap_proc = self.patch_nmap_proc.start()
        self.mock_nmap_proc().is_running = running_mock
        self.mock_nmap_proc().run = run_mock
        type(self.mock_nmap_proc()).progress = progress_mock
        self.mock_nmap_proc().stdout = "Hello Mock output"
        self.addCleanup(self.patch_nmap_proc.stop)
        self.addCleanup(mopen.stop)

    def tearDown(self):
        self.patch_nmap_proc.stop()

    @patch('os.path.isfile')
    def test_report_name_single_host(self, mock_isfile):
        mock_isfile.side_effect = [True, False, False, False]
        expected_calls = [
            call('fake/path/127.0.0.1.nmap'),
            call('fake/path/0-127.0.0.1.nmap'),
            call('fake/path/127.0.0.1.xml')
        ]
        sprocess = ScanProcess("fake/path")
        report = sprocess.run("127.0.0.1", "-p1-20", self.callback)
        self.assertEqual(b"127.0.0.1.xml", report.filename)
        mock_isfile.assert_has_calls(expected_calls)
        self.callback.assert_called_once_with(Status.SUCCESS)

    @patch('os.path.isfile')
    def test_report_name_network(self, mock_isfile):
        mock_isfile.return_value = False
        expected_call = [
            call('fake/path/127.0.0.1-24.xml')
        ]
        sprocess = ScanProcess("fake/path")
        report = sprocess.run("127.0.0.1/24", "-p1-20", self.callback)
        self.assertEqual(b"127.0.0.1-24.xml", report.filename)
        mock_isfile.assert_has_calls(expected_call)
        self.callback.assert_called_once_with(Status.SUCCESS)

    @patch('os.path.isfile')
    def test_report_name_network_existing(self, mock_isfile):
        mock_isfile.side_effect = [True, False, False, False]
        expected_calls = [
            call('fake/path/127.0.0.1-24.nmap'),
            call('fake/path/0-127.0.0.1-24.nmap'),
            call('fake/path/127.0.0.1-24.xml')
        ]
        sprocess = ScanProcess("fake/path")
        report = sprocess.run("127.0.0.1/24", "-p1-20", self.callback)
        self.assertIsNotNone(report)
        self.assertEqual(b"127.0.0.1-24.xml", report.filename)
        mock_isfile.assert_has_calls(expected_calls)
        self.callback.assert_called_once_with(Status.SUCCESS)


if __name__ == '__main__':
    unittest.main()



import hashlib
import hmac
import os
import struct
import unittest
from argparse import Namespace
from socket import timeout
from unittest.mock import MagicMock, call, mock_open, patch

import tests
from dscan.client import Agent
from dscan.models.scanner import Config, ScanProcess
from dscan.models.structures import (Auth, Command, ExitStatus, Ready, Report,
                                     Status)


class TestAgentHandler(unittest.TestCase):
    def setUp(self):
        options = Namespace(name='data', s='127.0.0.1', p=2040,
                            cmd='agent')
        self.challenge = b'4%x8p\x8d\xda\x04\xe5r\xfb\xc1Si8[' \
                         b'\xcb\x1a\x1c\x84\xf5\xb5%\x15[' \
                         b'\xea\x10\x96)!n\xe3\xadit\x0f\x15e\xc2\x06\xd1' \
                         b'\xd8\xb0' \
                         b'\xc5\x81\x87\xf2s\xe3\xd8\x95\xd1\x9c\xbdM\x8f' \
                         b'\x9c\xd5' \
                         b'\x14\xb3\x8e\xdd\x8eQ\xffw\x10Y8\xa5\xa5\x83\xf3' \
                         b'\xeeQ' \
                         b'\xa1\xfcOP\x9d\xd6\x80x\x80\x9eh\x11\xa7\xd7\xce' \
                         b'\xcf' \
                         b'.\xec\x01\x94S\xd4\x1d\x7f\xef\x83e\xe8\xfa\xf9' \
                         b'`\xfb' \
                         b'\xc6:SB\xeff\x15\r\xcb\xe9\xa4\xefO\x03i\xe9' \
                         b'\xefoMz\x8b'
        self.cfg = tests.create_config()
        self.patcher_makedirs = patch('os.makedirs')
        mos_isfile = patch('os.path.isfile')
        mos_isfile.return_value = True
        mos_isfile.start()
        self.mock_makedirs = self.patcher_makedirs.start()
        self.settings = Config(self.cfg, options)
        self.hmac_patch = patch.object(hmac, 'compare_digest',
                                       return_value=True)
        self.hmac_patch.start()
        self.patcher_defContext = patch('ssl.create_default_context')
        self.patcher_sslContext = patch('ssl.SSLContext', name="sslContext")
        self.patcher_socket = patch('socket.socket', name="socket")
        self.patcher_urandom = patch('os.urandom', name="urandom")
        self.mock_urandom = self.patcher_urandom.start()
        self.mock_urandom.return_value = self.challenge
        self.mock_create_ctx = self.patcher_defContext.start()
        self.mock_context = self.patcher_sslContext.start()
        self.mock_socket = self.patcher_socket.start()
        self.mock_create_ctx.return_value = self.mock_context
        self.mock_context.wrap_socket.return_value = self.mock_socket
        hmac_hash = hmac.new(self.settings.secret_key, self.challenge,
                             'sha512')
        self.digest_auth = hmac_hash.hexdigest().encode("utf-8")
        self.expected_calls_timeout = [
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.__bool__(),
            call.recv(128),
            call.sendall(Auth(self.digest_auth).pack()),
            call.recv(1),
            call.sendall(Ready(os.getuid(), "AAAAAA").pack()),
            call.recv(1),
            call.close(),
            call.close(),
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.__bool__(),
            call.recv(128),
            call.sendall(Auth(self.digest_auth).pack()),
            call.recv(1),
            call.sendall(Ready(os.getuid(), "AAAAAA").pack()),
            call.recv(1),
            call.close(),
            call.close(),
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.close(),
            call.close(),
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.close(),
            call.close(),
        ]
        self.addCleanup(self.patcher_makedirs.stop)
        self.addCleanup(self.hmac_patch.stop)
        self.addCleanup(self.patcher_defContext.stop)
        self.addCleanup(self.patcher_sslContext.stop)
        self.addCleanup(self.patcher_socket.stop)
        self.addCleanup(self.patcher_urandom.stop)
        self.addCleanup(mos_isfile.stop)

    def tearDown(self):
        self.patcher_makedirs.stop()
        self.hmac_patch.stop()
        self.patcher_defContext.stop()
        self.patcher_sslContext.stop()
        self.patcher_socket.stop()
        self.patcher_urandom.stop()

    def check_mock_calls_connect_disconnect(self):
        self.mock_context.wrap_socket.assert_called_once()
        self.mock_context.load_verify_locations.assert_called_once()
        self.mock_socket.connect.assert_called()
        self.mock_socket.close.assert_called()
        # Check call args
        self.mock_context.load_verify_locations.assert_called_once_with(
            'data/certfile.crt')
        self.mock_socket.connect.assert_called_with(
            ('127.0.0.1', 2040))

    def mock_server_responses(self, *args):
        buffer = tests.BufMock(*args)
        self.mock_socket.recv = buffer.read

    def test_connect(self):
        agent = Agent(self.settings)
        agent.start()
        self.check_mock_calls_connect_disconnect()

    def test_timeout_after_auth(self):
        mock_ex = MagicMock()
        mock_ex.side_effect = [struct.pack("<B", 1),
                               struct.pack("<128s", self.challenge),
                               struct.pack("<B", 0),
                               timeout(),
                               timeout(),
                               timeout(),
                               timeout()
                               ]
        expected = [
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.__bool__(),
            call.recv(128),
            call.sendall(Auth(self.digest_auth).pack()),
            call.recv(1),
            call.sendall(Ready(os.getuid(), "AAAAAA").pack()),
            call.recv(1),
            call.close(),
            call.close(),
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.close(),
            call.close(),
            call.connect(('127.0.0.1', 2040)),
            call.recv(1),
            call.close(),
            call.close(),
        ]
        self.mock_socket.recv = mock_ex
        with patch('random.choice') as mock_choice:
            mock_choice.return_value = "A"
            agent = Agent(self.settings)
            agent.start()
            self.mock_socket.assert_has_calls(expected)

    def test_reset_retry_count_timeout(self):
        mock_ex = MagicMock()
        mock_ex.side_effect = [struct.pack("<B", 1),
                               struct.pack("<128s", self.challenge),
                               struct.pack("<B", 0),
                               timeout(),
                               struct.pack("<B", 1),
                               struct.pack("<128s", self.challenge),
                               struct.pack("<B", 0),
                               timeout(),
                               timeout(),
                               timeout(),
                               ]
        self.mock_socket.recv = mock_ex
        with patch('random.choice') as mock_choice:
            mock_choice.return_value = "A"
            agent = Agent(self.settings)
            agent.start()
            self.mock_socket.assert_has_calls(self.expected_calls_timeout)

    @patch('os.getuid')
    def test_full(self, mgetuid):
        mgetuid.return_value = 0

        digest = hashlib.sha512(b"pickabu").hexdigest()
        data = "hello hello report mock\n"
        expected = Report(len(data), "fu.xml", digest)

        expected_calls = [
            call.connect(('127.0.0.1', 2040)),
            call.sendall(Auth(self.digest_auth).pack()),
            call.sendall(Ready(0, "AAAAAA").pack()),
            call.sendall(expected.pack()),
            call.sendall(data),
            call.sendall(data),
            call.sendall(Ready(0, "AAAAAA").pack()),
            call.close()
        ]

        self.mock_server_responses(Auth(self.challenge), struct.pack("<B", 0),
                                   Command("127.0.0.1", "-sV -Pn -p1-1000"),
                                   struct.pack("<B", 1), struct.pack("<B", 0))

        report_mock = mock_open(read_data=data)
        with patch('random.choice') as mock_choice:
            mock_choice.return_value = "A"
            with patch('builtins.open', report_mock):
                patcher = patch.object(ScanProcess, 'run',
                                       return_value=expected)
                patcher.start()
                agent = Agent(self.settings)
                agent.start()
                self.check_mock_calls_connect_disconnect()
                self.mock_socket.assert_has_calls(expected_calls,
                                                  any_order=True)
                patcher.stop()

    @patch('os.getuid')
    def test_wait(self, mgetuid):
        mgetuid.return_value = 0

        digest = hashlib.sha512(b"pickabu").hexdigest()
        data = "hello hello report mock\n"
        expected = Report(len(data), "fu.xml", digest)

        expected_calls = [
            call.connect(('127.0.0.1', 2040)),
            call.sendall(Auth(self.digest_auth).pack()),
            call.sendall(Ready(0, "AAAAAA").pack()),
            call.sendall(expected.pack()),
            call.sendall(data),
            call.sendall(data),
            call.sendall(Ready(0, "AAAAAA").pack()),
            call.close()
        ]

        self.mock_server_responses(Auth(self.challenge),
                                   struct.pack("<B", 0),
                                   ExitStatus(Status.UNFINISHED),
                                   Command("127.0.0.1", "-sV -Pn -p1-1000"),
                                   struct.pack("<B", 1), struct.pack("<B", 0))

        report_mock = mock_open(read_data=data)
        with patch('random.choice') as mock_choice:
            mock_choice.return_value = "A"
            with patch('builtins.open', report_mock):
                patcher = patch.object(ScanProcess, 'run',
                                       return_value=expected)
                patcher.start()
                agent = Agent(self.settings)
                agent.start()
                self.check_mock_calls_connect_disconnect()
                self.mock_socket.assert_has_calls(expected_calls,
                                                  any_order=True)
                patcher.stop()

    @patch('os.getuid')
    def test_full_unsuccessful_report(self, mgetuid):
        mgetuid.return_value = 0

        hmac_hash = hmac.new(self.settings.secret_key, self.challenge,
                             'sha512')
        digest_auth = hmac_hash.hexdigest().encode("utf-8")
        digest = hashlib.sha512(b"pickabu").hexdigest()
        data = "hello hello report mock\n"
        expected = Report(len(data), "fu.xml", digest)

        expected_calls = [
            call.connect(('127.0.0.1', 2040)),
            call.sendall(Auth(digest_auth).pack()),
            call.sendall(Ready(0, "AAAAAA").pack()),
            call.sendall(expected.pack()),
            call.sendall(data),
            call.sendall(data),
            call.sendall(data),
            call.close()
        ]

        self.mock_server_responses(Auth(self.challenge), struct.pack("<B", 0),
                                   Command("127.0.0.1", "-sV -Pn -p1-1000"),
                                   struct.pack("<B", 1), struct.pack("<B", 1),
                                   struct.pack("<B", 1))

        report_mock = mock_open(read_data=data)
        with patch('random.choice') as mock_choice:
            mock_choice.return_value = "A"
            with patch('builtins.open', report_mock):
                patcher = patch.object(ScanProcess, 'run',
                                       return_value=expected)
                patcher.start()
                agent = Agent(self.settings)
                agent.start()
                self.check_mock_calls_connect_disconnect()
                self.mock_socket.assert_has_calls(expected_calls,
                                                  any_order=True)
                patcher.stop()


if __name__ == '__main__':
    unittest.main()



from setuptools import find_packages, setup

with open('README.md', encoding='utf-8') as f:
    readme_text = f.read()

name = 'dscan-4E656F'
version = '0.1.6'
release = '0.1.6'

setup(
    name=name,
    version=version,
    description=(
        'Wrapper around nmap allow nmap scans to be distributed',
        'With resume, and address optimization.'
    ),
    long_description=readme_text,
    long_description_content_type='text/markdown',
    license="GPL version 3",
    author='0x4E0x650x6F',
    author_email='tiago.alexand@gmail.com',
    url='https://github.com/0x4E0x650x6F/dscan',
    tests_requires=[
        "autopep8", "flake8", "isort", "sphinx",
    ],
    install_requires=[
        'python-libnmap'
    ],
    scripts=['bin/dscan'],
    package_data={
          name: ['data/agent.conf', 'data/dscan.conf'],
       },
    packages=find_packages(include=['dscan', 'dscan.*'], exclude='tests'),
    include_package_data=True,
    python_requires='>=3.6',
    classifiers=[
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Topic :: System :: Networking",
        ],
)



"""
parsers.py
parsers models, input ip address list collapse, or
scanner results to parse.
"""
import fnmatch
import ipaddress
import os
import argparse
from libnmap.parser import NmapParser, NmapParserException

from dscan import log


def parse_args():
    """
    Used by main to parse the user arguments.

    :return: argparse instance.
    :rtype: `argparse.ArgumentParser`
    """
    parser = argparse.ArgumentParser(prog='Distributed scanner')
    parser.add_argument('--name', type=str, required=True)

    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True
    parser_server = subparsers.add_parser('srv')
    parser_server.add_argument('--config', required=True)
    parser_server.add_argument('-b', default='0.0.0.0')
    parser_server.add_argument('-p', type=int, default=2040)
    parser_server.add_argument('targets', type=argparse.FileType('rt'))
    parser_agent = subparsers.add_parser('agent')
    parser_agent.add_argument('--config', required=True)
    parser_agent.add_argument('-s', required=True)
    parser_agent.add_argument('-p', type=int, default=2040)
    parser_config = subparsers.add_parser('config')
    parser_config.add_argument("-email", type=str, required=True)
    parser_config.add_argument("-cn", type=str, required=True)
    parser_config.add_argument("-c", type=str, required=True)
    parser_config.add_argument("-l", type=str, required=True)
    parser_config.add_argument("-st", type=str, required=True)
    parser_config.add_argument("-o", type=str, required=True)
    parser_config.add_argument("-ou", type=str, required=True)
    parser_config.add_argument("-days", type=int, required=True)
    return parser


class ReportsParser:
    """
    XML Nmap results parser.
    """

    def __init__(self, reports_path, pattern):
        """
        :param reports_path: path where the reports are stored
        :param pattern: pattern `fnmatch` to find valid files to extract
        the results from.
        """
        self.path = reports_path
        self.pattern = pattern

    def hosts_up(self):
        """
        :return: list of hosts up.
        :rtype: `list`
        """
        hosts_up = []
        for host in self.__walk():
            if host.is_up():
                hosts_up.append(host.ipv4)
        return hosts_up

    def __walk(self):
        """
        information.
        :yield: A list with the filtered values
        :rtype: `list`
        """
        for report in os.scandir(self.path):
            if fnmatch.fnmatch(report.name, self.pattern):
                try:
                    nmap_report = NmapParser.parse_fromfile(report.path)
                    yield from nmap_report.hosts
                except NmapParserException as ex:
                    log.error(f"Error parsing {report} - {ex}")


class TargetOptimization:
    """
    This class takes lists of hosts or networks, and attempts to optimize
        them by either split big cidr like /8 /16 in /24 or in rage format
        192.168.10.1-4.
    """

    def __init__(self, fpath, cidr="/24"):
        self.cidr = cidr
        self.fpath = fpath

    def save(self, targets):
        """
        Takes a list of targets to optimize and saves it in the workspace path.

        :param targets: `list` of targets (`str` and top optimize.
        :type: targets: `list` of `str`
        """
        assert targets, "Empty target list"
        ips = []

        with open(self.fpath, 'wt') as qfile:
            for target in targets:
                try:
                    if "/" in target:
                        net = ipaddress.ip_network(target.strip())
                        if net.prefixlen < 24:
                            subs = map(lambda n: f"{n.with_prefixlen}\n",
                                       net.subnets(new_prefix=24))
                            qfile.writelines(subs)
                        else:
                            qfile.write(f"{net.with_prefixlen}\n")
                    else:
                        ips.append(ipaddress.ip_address(target.strip()))
                except (TypeError, ValueError):
                    log.error(f"Error optimizing target: {target}")

            # sorting the ip addresses.
            ips.sort(key=ipaddress.get_mixed_type_key)
            # find consecutive ip address ranges.
            if ips:
                for first, last in ipaddress._find_address_range(ips):
                    ip_range = list(ipaddress.summarize_address_range(first,
                                                                      last))
                    # if the number of ranges is more than one network in cidr
                    # format then the glob format x.x.x.x-y is more efficient,
                    # since nmap supports this format.
                    if len(ip_range) > 1:
                        qfile.write(f"{first}-{last.exploded.split('.')[3]}\n")
                    else:
                        qfile.write(f"{ip_range.pop().with_prefixlen}\n")